import pytest

import salt.modules.saltutil
import salt.states.saltmod as saltmod
import salt.utils.event
import salt.utils.jid
import salt.utils.state
from tests.support.mock import MagicMock, create_autospec, patch


@pytest.fixture
def configure_loader_modules(minion_opts):
    return {
        saltmod: {
            "__opts__": minion_opts,
            "__salt__": {"saltutil.cmd": MagicMock()},
            "__utils__": {"state.check_result": salt.utils.state.check_result},
        },
    }


@pytest.fixture
def fake_cmd():
    _fake_cmd = create_autospec(salt.modules.saltutil.cmd)
    with patch.dict(saltmod.__salt__, {"saltutil.cmd": _fake_cmd}):
        yield _fake_cmd


@pytest.mark.parametrize(
    "exclude",
    [True, False],
)
def test_exclude_parameter_gets_passed(exclude, fake_cmd):
    """
    Smoke test for for salt.states.statemod.state().  Ensures that we
    don't take an exception if optional parameters are not specified in
    __opts__ or __env__.
    """
    args = ("webserver_setup", "webserver2")
    expected_exclude = exclude
    kwargs = {
        "tgt_type": "glob",
        "exclude": expected_exclude,
        "highstate": True,
    }

    saltmod.state(*args, **kwargs)

    call = fake_cmd.call_args[1]
    assert call["kwarg"]["exclude"] == expected_exclude


def test_exclude_parameter_is_not_passed_if_not_provided(fake_cmd):
    # Make sure we don't barf on existing behavior
    args = ("webserver_setup", "webserver2")
    kwargs_without_exclude = {
        "tgt_type": "glob",
        "highstate": True,
    }

    saltmod.state(*args, **kwargs_without_exclude)

    call = fake_cmd.call_args[1]
    assert "exclude" not in call["kwarg"]


def test_state_smoke_test():
    """
    Smoke test for for salt.states.statemod.state().  Ensures that we
    don't take an exception if optional parameters are not specified in
    __opts__ or __env__.
    """
    args = ("webserver_setup", "webserver2")
    kwargs = {
        "tgt_type": "glob",
        "fail_minions": None,
        "pillar": None,
        "top": None,
        "batch": None,
        "orchestration_jid": None,
        "sls": "vroom",
        "queue": False,
        "concurrent": False,
        "highstate": None,
        "expr_form": None,
        "ret": "",
        "ssh": False,
        "timeout": None,
        "test": False,
        "allow_fail": 0,
        "saltenv": None,
        "expect_minions": False,
    }
    with patch.dict(saltmod.__opts__, {"id": "webserver2"}):
        ret = saltmod.state(*args, **kwargs)
    expected = {
        "comment": "States ran successfully.",
        "changes": {},
        "name": "webserver_setup",
        "result": True,
    }
    assert ret == expected


def test_state():
    """
    Test to invoke a state run on a given target
    """
    name = "state"
    tgt = "minion1"

    expected = {
        "name": name,
        "changes": {},
        "result": False,
        "comment": "No highstate or sls specified, no execution made",
    }
    ret = saltmod.state(name, tgt)
    assert ret == expected

    expected.update({"comment": "Must pass in boolean for value of 'concurrent'"})
    ret = saltmod.state(name, tgt, highstate=True, concurrent="a")
    assert ret == expected

    expected.update(
        {
            "result": True,
            "comment": "States ran successfully.",
        }
    )
    with patch.dict(saltmod.__opts__, {"test": True}):
        ret = saltmod.state(name, tgt, highstate=True)
    assert ret == expected

    silver_ret = {
        "test_|-notify_me_|-this is a name_|-show_notification": {
            "comment": "Notify me",
            "name": "this is a name",
            "start_time": "10:43:41.487565",
            "result": True,
            "duration": 0.35,
            "__run_num__": 0,
            "__sls__": "demo",
            "changes": {"foo": "bar"},
            "__id__": "notify_me",
        }
    }
    expected.update(
        {
            "comment": "States ran successfully. Updating silver.",
            "result": None,
            "__jid__": "20170406104341210934",
            "changes": {
                "out": "highstate",
                "ret": {"silver": silver_ret},
            },
        }
    )
    with patch.dict(saltmod.__opts__, {"test": True}):
        mock = MagicMock(
            return_value={
                "silver": {
                    "jid": "20170406104341210934",
                    "retcode": 0,
                    "ret": silver_ret,
                    "out": "highstate",
                }
            }
        )
        with patch.dict(saltmod.__salt__, {"saltutil.cmd": mock}):
            ret = saltmod.state(name, tgt, highstate=True)
            assert ret == expected
        mock.assert_called_once()

    expected.update(
        {
            "comment": "States ran successfully. No changes made to silver.",
            "result": True,
            "__jid__": "20170406104341210934",
            "changes": {},
        }
    )
    with patch.dict(saltmod.__opts__, {"test": False}):
        mock = MagicMock(
            return_value={
                "silver": {
                    "jid": "20170406104341210934",
                    "retcode": 0,
                    "ret": {
                        "test_|-notify_me_|-this is a name_|-show_notification": {
                            "comment": "Notify me",
                            "name": "this is a name",
                            "start_time": "10:43:41.487565",
                            "result": True,
                            "duration": 0.35,
                            "__run_num__": 0,
                            "__sls__": "demo",
                            "changes": {},
                            "__id__": "notify_me",
                        }
                    },
                    "out": "highstate",
                }
            }
        )
        with patch.dict(saltmod.__salt__, {"saltutil.cmd": mock}):
            ret = saltmod.state(name, tgt, highstate=True)
            assert ret == expected
        mock.assert_called_once()

        with patch.dict(saltmod.__salt__, {"saltutil.cmd": mock}):
            ret = saltmod.state(name, tgt, top="the-top")
            assert "arg" in mock.call_args.kwargs
            assert "the-top" in mock.call_args.kwargs["arg"]

        for pass_kw in ("ret_config", "ret_kwargs", "batch", "subset"):
            with patch.dict(saltmod.__salt__, {"saltutil.cmd": mock}):
                kwargs = {pass_kw: f"{pass_kw}_value"}
                ret = saltmod.state(name, tgt, highstate=True, **{pass_kw: kwargs})
            assert pass_kw in mock.call_args.kwargs
            if pass_kw == "batch":
                assert mock.call_args.kwargs[pass_kw] == str(kwargs)
            else:
                assert mock.call_args.kwargs[pass_kw] == kwargs
            assert ret == expected

        for pass_kw in ("pillar", "pillarenv", "saltenv"):
            with patch.dict(saltmod.__salt__, {"saltutil.cmd": mock}):
                kwargs = {pass_kw: f"{pass_kw}_value"}
                ret = saltmod.state(name, tgt, highstate=True, **{pass_kw: kwargs})
            assert "kwarg" in mock.call_args.kwargs
            assert pass_kw in mock.call_args.kwargs["kwarg"]
            assert mock.call_args.kwargs["kwarg"][pass_kw] == kwargs
            assert ret == expected

    test_batch_return = {
        "minion1": {
            "ret": {
                "test_|-notify_me_|-this is a name_|-show_notification": {
                    "comment": "Notify me",
                    "name": "this is a name",
                    "start_time": "10:43:41.487565",
                    "result": True,
                    "duration": 0.35,
                    "__run_num__": 0,
                    "__sls__": "demo",
                    "changes": {},
                    "__id__": "notify_me",
                },
                "retcode": 0,
            },
            "out": "highstate",
        },
        "minion2": {
            "ret": {
                "test_|-notify_me_|-this is a name_|-show_notification": {
                    "comment": "Notify me",
                    "name": "this is a name",
                    "start_time": "10:43:41.487565",
                    "result": True,
                    "duration": 0.35,
                    "__run_num__": 0,
                    "__sls__": "demo",
                    "changes": {},
                    "__id__": "notify_me",
                },
                "retcode": 0,
            },
            "out": "highstate",
        },
        "minion3": {
            "ret": {
                "test_|-notify_me_|-this is a name_|-show_notification": {
                    "comment": "Notify me",
                    "name": "this is a name",
                    "start_time": "10:43:41.487565",
                    "result": True,
                    "duration": 0.35,
                    "__run_num__": 0,
                    "__sls__": "demo",
                    "changes": {},
                    "__id__": "notify_me",
                },
                "retcode": 0,
            },
            "out": "highstate",
        },
    }
    expected.update(
        {
            "comment": (
                "States ran successfully. No changes made to minion1, minion3,"
                " minion2."
            )
        }
    )
    del expected["__jid__"]
    with patch.dict(saltmod.__opts__, {"test": False}):
        with patch.dict(
            saltmod.__salt__,
            {"saltutil.cmd": MagicMock(return_value=test_batch_return)},
        ):
            state_run = saltmod.state(name, tgt, highstate=True)

            # Test return without checking the comment contents. Comments are tested later.
            comment = state_run.pop("comment")
            expected.pop("comment")
            assert state_run == expected

            # Check the comment contents in a non-order specific way (ordering fails sometimes on PY3)
            assert "States ran successfully. No changes made to" in comment
            for minion in ["minion1", "minion2", "minion3"]:
                assert minion in comment


def test_state_masterless():
    """
    Test to invoke a state run masterless
    """
    name = "state"
    minion_id = "masterless-minion"

    expected = {
        "name": name,
        "changes": {},
        "comment": f"States ran successfully. No changes made to {minion_id}.",
        "result": True,
    }
    with patch.dict(
        saltmod.__opts__,
        {"test": False, "__role": "minion", "file_client": "local", "id": minion_id},
    ):
        mock = MagicMock(
            return_value={
                minion_id: {
                    "jid": "20170406104341210934",
                    "retcode": 0,
                    "ret": {
                        "test_|-notify_me_|-this is a name_|-show_notification": {
                            "comment": "Notify me",
                            "name": "this is a name",
                            "start_time": "10:43:41.487565",
                            "result": True,
                            "duration": 0.35,
                            "__run_num__": 0,
                            "__sls__": "demo",
                            "changes": {},
                            "__id__": "notify_me",
                        }
                    },
                    "out": "highstate",
                }
            }
        )
        with patch.dict(saltmod.__salt__, {"state.highstate": mock}):
            ret = saltmod.state(name, minion_id, highstate=True)
            assert ret == expected
            mock.assert_called_once()
        with patch.dict(saltmod.__salt__, {"state.top": mock}):
            ret = saltmod.state(name, minion_id, top="the-top")
            assert ret == expected
            assert "topfn" in mock.call_args.kwargs
            assert mock.call_args.kwargs["topfn"] == "the-top"
        with patch.dict(saltmod.__salt__, {"state.sls": mock}):
            ret = saltmod.state(name, minion_id, sls="the-sls")
            assert ret == expected
            assert "mods" in mock.call_args.kwargs
            assert mock.call_args.kwargs["mods"] == "the-sls"
        with patch.dict(saltmod.__salt__, {"state.sls": mock}):
            ret = saltmod.state(name, minion_id, sls=["the-sls-1", "the-sls-2"])
            assert ret == expected
            assert "mods" in mock.call_args.kwargs
            assert mock.call_args.kwargs["mods"] == "the-sls-1,the-sls-2"


def test_state_failhard():

    name = "state"
    tgt = "minion1"

    expected = {
        "name": name,
        "changes": {},
        "comment": "States ran successfully. No changes made to silver.",
        "result": True,
        "__jid__": "20170406104341210934",
    }
    with patch.dict(saltmod.__opts__, {"test": False}):
        mock = MagicMock(
            return_value={
                "silver": {
                    "jid": "20170406104341210934",
                    "retcode": 0,
                    "ret": {
                        "test_|-notify_me_|-this is a name_|-show_notification": {
                            "comment": "Notify me",
                            "name": "this is a name",
                            "start_time": "10:43:41.487565",
                            "result": True,
                            "duration": 0.35,
                            "__run_num__": 0,
                            "__sls__": "demo",
                            "changes": {},
                            "__id__": "notify_me",
                        }
                    },
                    "out": "highstate",
                }
            }
        )
        with patch.dict(saltmod.__salt__, {"saltutil.cmd": mock}):
            ret = saltmod.state(name, tgt, highstate=True, failhard=True)
            assert ret == expected
        mock.assert_called_once()
        assert "failhard" in mock.call_args.kwargs
        assert mock.call_args.kwargs["failhard"] is True

    with patch.dict(saltmod.__opts__, {"test": False, "failhard": True}):
        mock = MagicMock(
            return_value={
                "silver": {
                    "jid": "20170406104341210934",
                    "retcode": 0,
                    "ret": {
                        "test_|-notify_me_|-this is a name_|-show_notification": {
                            "comment": "Notify me",
                            "name": "this is a name",
                            "start_time": "10:43:41.487565",
                            "result": True,
                            "duration": 0.35,
                            "__run_num__": 0,
                            "__sls__": "demo",
                            "changes": {},
                            "__id__": "notify_me",
                        }
                    },
                    "out": "highstate",
                }
            }
        )
        with patch.dict(saltmod.__salt__, {"saltutil.cmd": mock}):
            ret = saltmod.state(name, tgt, highstate=True)
            assert ret == expected
        mock.assert_called_once()
        assert "failhard" in mock.call_args.kwargs
        assert mock.call_args.kwargs["failhard"] is True


def test_state_no_returns():

    name = "state"
    tgt = "minion1"

    expected = {
        "name": name,
        "changes": {},
        "result": False,
        "comment": "No minions returned",
    }
    with patch.dict(saltmod.__opts__, {"test": False}):
        mock = MagicMock(return_value={})
        with patch.dict(saltmod.__salt__, {"saltutil.cmd": mock}):
            ret = saltmod.state(name, tgt, highstate=True)
            assert ret == expected
        mock.assert_called_once()


def test_state_failed_and_expected_minions():

    name = "state"
    tgt = "minion1"

    expected = {
        "name": name,
        "changes": {"out": "highstate", "ret": {"silver": False}},
        "comment": "Run failed on minions: silver",
        "result": False,
        "__jid__": "20170406104341210934",
    }
    with patch.dict(saltmod.__opts__, {"test": False}):
        mock = MagicMock(
            return_value={
                "silver": {
                    "jid": "20170406104341210934",
                    "retcode": 0,
                    "ret": {
                        "test_|-notify_me_|-this is a name_|-show_notification": {
                            "comment": "Notify me",
                            "name": "this is a name",
                            "start_time": "10:43:41.487565",
                            "result": True,
                            "duration": 0.35,
                            "__run_num__": 0,
                            "__sls__": "demo",
                            "changes": {},
                            "__id__": "notify_me",
                        }
                    },
                    "failed": True,
                    "out": "highstate",
                },
                "gold": {
                    "jid": "20170406104341210934",
                    "retcode": 0,
                    "ret": {
                        "test_|-notify_me_|-this is a name_|-show_notification": {
                            "comment": "Notify me",
                            "name": "this is a name",
                            "start_time": "10:43:41.487565",
                            "result": True,
                            "duration": 0.35,
                            "__run_num__": 0,
                            "__sls__": "demo",
                            "changes": {},
                            "__id__": "notify_me",
                        }
                    },
                    "out": "highstate",
                },
            }
        )
        with patch.dict(saltmod.__salt__, {"saltutil.cmd": mock}):
            ret = saltmod.state(name, tgt, highstate=True)
            assert ret == expected
        mock.assert_called_once()

    expected.update(
        {
            "changes": {
                "out": "highstate",
                "ret": {"bronze": False, "charcoal": False, "silver": False},
            },
            "comment": "Run failed on minions: silver, bronze",
        }
    )
    with patch.dict(saltmod.__opts__, {"test": False}):
        mock = MagicMock(
            return_value={
                "charcoal": {
                    "jid": "20170406104341210934",
                    "retcode": 0,
                    "ret": {
                        "test_|-notify_me_|-this is a name_|-show_notification": {
                            "comment": "Notify me",
                            "name": "this is a name",
                            "start_time": "10:43:41.487565",
                            "result": True,
                            "duration": 0.35,
                            "__run_num__": 0,
                            "__sls__": "demo",
                            "changes": {},
                            "__id__": "notify_me",
                        }
                    },
                    "failed": True,
                    "out": "highstate",
                },
                "bronze": {
                    "jid": "20170406104341210934",
                    "retcode": 0,
                    "ret": {
                        "test_|-notify_me_|-this is a name_|-show_notification": {
                            "comment": "Notify me",
                            "name": "this is a name",
                            "start_time": "10:43:41.487565",
                            "result": True,
                            "duration": 0.35,
                            "__run_num__": 0,
                            "__sls__": "demo",
                            "changes": {},
                            "__id__": "notify_me",
                        }
                    },
                    "failed": True,
                    "out": "highstate",
                },
                "silver": {
                    "jid": "20170406104341210934",
                    "retcode": 0,
                    "ret": {
                        "test_|-notify_me_|-this is a name_|-show_notification": {
                            "comment": "Notify me",
                            "name": "this is a name",
                            "start_time": "10:43:41.487565",
                            "result": True,
                            "duration": 0.35,
                            "__run_num__": 0,
                            "__sls__": "demo",
                            "changes": {},
                            "__id__": "notify_me",
                        }
                    },
                    "failed": True,
                    "out": "highstate",
                },
                "gold": {
                    "jid": "20170406104341210934",
                    "retcode": 0,
                    "ret": {
                        "test_|-notify_me_|-this is a name_|-show_notification": {
                            "comment": "Notify me",
                            "name": "this is a name",
                            "start_time": "10:43:41.487565",
                            "result": True,
                            "duration": 0.35,
                            "__run_num__": 0,
                            "__sls__": "demo",
                            "changes": {},
                            "__id__": "notify_me",
                        }
                    },
                    "out": "highstate",
                },
            }
        )
        with patch.dict(saltmod.__salt__, {"saltutil.cmd": mock}):
            ret = saltmod.state(name, tgt, highstate=True, fail_minions="charcoal")
            ret_comment = ret.pop("comment")
            expected.pop("comment")
            assert ret == expected
            # The order can be different, hence asserting like this
            assert "Run failed on minions: " in ret_comment
            assert "silver" in ret_comment
            assert "bronze" in ret_comment
        mock.assert_called_once()

    expected.update(
        {
            "changes": {
                "out": "highstate",
                "ret": {"bronze": False, "charcoal": False, "silver": False},
            },
            "comment": "Run failed on minions: silver",
        }
    )
    with patch.dict(saltmod.__opts__, {"test": False}):
        mock = MagicMock(
            return_value={
                "charcoal": {
                    "jid": "20170406104341210934",
                    "retcode": 0,
                    "ret": {
                        "test_|-notify_me_|-this is a name_|-show_notification": {
                            "comment": "Notify me",
                            "name": "this is a name",
                            "start_time": "10:43:41.487565",
                            "result": True,
                            "duration": 0.35,
                            "__run_num__": 0,
                            "__sls__": "demo",
                            "changes": {},
                            "__id__": "notify_me",
                        }
                    },
                    "failed": True,
                    "out": "highstate",
                },
                "bronze": {
                    "jid": "20170406104341210934",
                    "retcode": 0,
                    "ret": {
                        "test_|-notify_me_|-this is a name_|-show_notification": {
                            "comment": "Notify me",
                            "name": "this is a name",
                            "start_time": "10:43:41.487565",
                            "result": True,
                            "duration": 0.35,
                            "__run_num__": 0,
                            "__sls__": "demo",
                            "changes": {},
                            "__id__": "notify_me",
                        }
                    },
                    "failed": True,
                    "out": "highstate",
                },
                "silver": {
                    "jid": "20170406104341210934",
                    "retcode": 0,
                    "ret": {
                        "test_|-notify_me_|-this is a name_|-show_notification": {
                            "comment": "Notify me",
                            "name": "this is a name",
                            "start_time": "10:43:41.487565",
                            "result": True,
                            "duration": 0.35,
                            "__run_num__": 0,
                            "__sls__": "demo",
                            "changes": {},
                            "__id__": "notify_me",
                        }
                    },
                    "failed": True,
                    "out": "highstate",
                },
                "gold": {
                    "jid": "20170406104341210934",
                    "retcode": 0,
                    "ret": {
                        "test_|-notify_me_|-this is a name_|-show_notification": {
                            "comment": "Notify me",
                            "name": "this is a name",
                            "start_time": "10:43:41.487565",
                            "result": True,
                            "duration": 0.35,
                            "__run_num__": 0,
                            "__sls__": "demo",
                            "changes": {},
                            "__id__": "notify_me",
                        }
                    },
                    "out": "highstate",
                },
            }
        )
        with patch.dict(saltmod.__salt__, {"saltutil.cmd": mock}):
            ret = saltmod.state(
                name, tgt, highstate=True, fail_minions="bronze,charcoal"
            )
            assert ret == expected
        mock.assert_called_once()
        with patch.dict(saltmod.__salt__, {"saltutil.cmd": mock}):
            ret = saltmod.state(
                name, tgt, highstate=True, fail_minions=["bronze", "charcoal"]
            )
            assert ret == expected

    expected.pop("__jid__")
    expected.update(
        {
            "result": True,
            "changes": {},
            "comment": "States ran successfully.",
            "warnings": [
                "'fail_minions' needs to be a list or a comma separated string. Ignored.",
            ],
        }
    )
    ret = saltmod.state(name, tgt, highstate=True, fail_minions={})
    assert ret == expected


def test_state_allow_fail():

    name = "state"
    tgt = "minion1"

    expected = {
        "name": name,
        "changes": {"out": "highstate", "ret": {"silver": False}},
        "comment": "States ran successfully. Updating silver. No changes made to gold.",
        "result": True,
        "__jid__": "20170406104341210934",
    }
    with patch.dict(saltmod.__opts__, {"test": False}):
        mock = MagicMock(
            return_value={
                "silver": {
                    "jid": "20170406104341210934",
                    "retcode": 0,
                    "ret": {
                        "test_|-notify_me_|-this is a name_|-show_notification": {
                            "comment": "Notify me",
                            "name": "this is a name",
                            "start_time": "10:43:41.487565",
                            "result": True,
                            "duration": 0.35,
                            "__run_num__": 0,
                            "__sls__": "demo",
                            "changes": {},
                            "__id__": "notify_me",
                        }
                    },
                    "failed": True,
                    "out": "highstate",
                },
                "gold": {
                    "jid": "20170406104341210934",
                    "retcode": 0,
                    "ret": {
                        "test_|-notify_me_|-this is a name_|-show_notification": {
                            "comment": "Notify me",
                            "name": "this is a name",
                            "start_time": "10:43:41.487565",
                            "result": True,
                            "duration": 0.35,
                            "__run_num__": 0,
                            "__sls__": "demo",
                            "changes": {},
                            "__id__": "notify_me",
                        }
                    },
                    "out": "highstate",
                },
            }
        )
        with patch.dict(saltmod.__salt__, {"saltutil.cmd": mock}):
            ret = saltmod.state(name, tgt, highstate=True, allow_fail=1)
            assert ret == expected
        mock.assert_called_once()

    gold_ret = {
        "test_|-notify_me_|-this is a name_|-show_notification": {
            "comment": "Notify me",
            "name": "this is a name",
            "start_time": "10:43:41.487565",
            "result": True,
            "duration": 0.35,
            "__run_num__": 0,
            "__sls__": "demo",
            "changes": {"foo": "bar"},
            "__id__": "notify_me",
        }
    }

    expected["changes"]["ret"]["gold"] = gold_ret

    with patch.dict(saltmod.__opts__, {"test": False}):
        mock = MagicMock(
            return_value={
                "silver": {
                    "jid": "20170406104341210934",
                    "retcode": 0,
                    "ret": {
                        "test_|-notify_me_|-this is a name_|-show_notification": {
                            "comment": "Notify me",
                            "name": "this is a name",
                            "start_time": "10:43:41.487565",
                            "result": True,
                            "duration": 0.35,
                            "__run_num__": 0,
                            "__sls__": "demo",
                            "changes": {},
                            "__id__": "notify_me",
                        }
                    },
                    "failed": True,
                    "out": "highstate",
                },
                "gold": {
                    "jid": "20170406104341210934",
                    "retcode": 0,
                    "ret": gold_ret,
                    "out": "highstate",
                },
            }
        )
        with patch.dict(saltmod.__salt__, {"saltutil.cmd": mock}):
            ret = saltmod.state(name, tgt, highstate=True, allow_fail=1)
            ret_comment = ret.pop("comment")
            expected.pop("comment")
            assert ret == expected
            # The order can be different, hence asserting like this
            assert "States ran successfully. Updating " in ret_comment
            assert "silver" in ret_comment
            assert "gold" in ret_comment
        mock.assert_called_once()

    expected.update(
        {
            "changes": {
                "out": "highstate",
                "ret": {"bronze": False, "charcoal": False, "silver": False},
            },
            "comment": "Run failed on minions: silver, bronze",
            "result": False,
        }
    )
    with patch.dict(saltmod.__opts__, {"test": False}):
        mock = MagicMock(
            return_value={
                "charcoal": {
                    "jid": "20170406104341210934",
                    "retcode": 0,
                    "ret": {
                        "test_|-notify_me_|-this is a name_|-show_notification": {
                            "comment": "Notify me",
                            "name": "this is a name",
                            "start_time": "10:43:41.487565",
                            "result": True,
                            "duration": 0.35,
                            "__run_num__": 0,
                            "__sls__": "demo",
                            "changes": {},
                            "__id__": "notify_me",
                        }
                    },
                    "failed": True,
                    "out": "highstate",
                },
                "bronze": {
                    "jid": "20170406104341210934",
                    "retcode": 0,
                    "ret": {
                        "test_|-notify_me_|-this is a name_|-show_notification": {
                            "comment": "Notify me",
                            "name": "this is a name",
                            "start_time": "10:43:41.487565",
                            "result": True,
                            "duration": 0.35,
                            "__run_num__": 0,
                            "__sls__": "demo",
                            "changes": {},
                            "__id__": "notify_me",
                        }
                    },
                    "failed": True,
                    "out": "highstate",
                },
                "silver": {
                    "jid": "20170406104341210934",
                    "retcode": 0,
                    "ret": {
                        "test_|-notify_me_|-this is a name_|-show_notification": {
                            "comment": "Notify me",
                            "name": "this is a name",
                            "start_time": "10:43:41.487565",
                            "result": True,
                            "duration": 0.35,
                            "__run_num__": 0,
                            "__sls__": "demo",
                            "changes": {},
                            "__id__": "notify_me",
                        }
                    },
                    "failed": True,
                    "out": "highstate",
                },
                "gold": {
                    "jid": "20170406104341210934",
                    "retcode": 0,
                    "ret": {
                        "test_|-notify_me_|-this is a name_|-show_notification": {
                            "comment": "Notify me",
                            "name": "this is a name",
                            "start_time": "10:43:41.487565",
                            "result": True,
                            "duration": 0.35,
                            "__run_num__": 0,
                            "__sls__": "demo",
                            "changes": {},
                            "__id__": "notify_me",
                        }
                    },
                    "out": "highstate",
                },
            }
        )
        with patch.dict(saltmod.__salt__, {"saltutil.cmd": mock}):
            ret = saltmod.state(name, tgt, highstate=True, allow_fail=1)
            ret_comment = ret.pop("comment")
            expected.pop("comment")
            assert ret == expected
            # The order can be different, hence asserting like this
            assert "Run failed on minions: " in ret_comment
            assert "silver" in ret_comment
            assert "bronze" in ret_comment
        mock.assert_called_once()

    expected = {
        "name": name,
        "changes": {},
        "result": False,
        "comment": "Passed invalid value for 'allow_fail', must be an int",
    }
    ret = saltmod.state(name, tgt, allow_fail="a")
    assert ret == expected


def test_roster():
    """
    Test saltmod state passes roster to saltutil.cmd
    """
    cmd_mock = MagicMock()
    with patch.dict(saltmod.__salt__, {"saltutil.cmd": cmd_mock}):
        ret = saltmod.state(
            "state.sls", tgt="*", ssh=True, highstate=True, roster="my_roster"
        )
    assert "roster" in cmd_mock.call_args.kwargs
    assert cmd_mock.call_args.kwargs["roster"] == "my_roster"



"""
    Test for the salt.modules.tuned
"""

import pytest

from salt.modules import tuned
from tests.support.mock import MagicMock, patch


@pytest.fixture
def configure_loader_modules():
    return {tuned: {}}


def test_v_241():
    """
    Test the list_ function for older tuned-adm (v2.4.1)
    as shipped with CentOS-6
    """
    tuned_list = """Available profiles:
- throughput-performance
- virtual-guest
- latency-performance
- laptop-battery-powersave
- laptop-ac-powersave
- virtual-host
- desktop-powersave
- server-powersave
- spindown-disk
- sap
- enterprise-storage
- default
Current active profile: throughput-performance"""
    mock_cmd = MagicMock(return_value=tuned_list)
    with patch.dict(tuned.__salt__, {"cmd.run": mock_cmd}):
        assert tuned.list_() == [
            "throughput-performance",
            "virtual-guest",
            "latency-performance",
            "laptop-battery-powersave",
            "laptop-ac-powersave",
            "virtual-host",
            "desktop-powersave",
            "server-powersave",
            "spindown-disk",
            "sap",
            "enterprise-storage",
            "default",
        ]


def test_v_271():
    """
    Test the list_ function for newer tuned-adm (v2.7.1)
    as shipped with CentOS-7
    """
    tuned_list = """Available profiles:
- balanced                    - General non-specialized tuned profile
- desktop                     - Optmize for the desktop use-case
- latency-performance         - Optimize for deterministic performance
- network-latency             - Optimize for deterministic performance
- network-throughput          - Optimize for streaming network throughput.
- powersave                   - Optimize for low power-consumption
- throughput-performance      - Broadly applicable tuning that provides--
- virtual-guest               - Optimize for running inside a virtual-guest.
- virtual-host                - Optimize for running KVM guests
Current active profile: virtual-guest
"""
    mock_cmd = MagicMock(return_value=tuned_list)
    with patch.dict(tuned.__salt__, {"cmd.run": mock_cmd}):
        assert tuned.list_() == [
            "balanced",
            "desktop",
            "latency-performance",
            "network-latency",
            "network-throughput",
            "powersave",
            "throughput-performance",
            "virtual-guest",
            "virtual-host",
        ]


def test_v_2110_with_warnings():
    """
    Test the list_ function for newer tuned-adm (v2.11.0)
    as shipped with CentOS-7.8 when warnings are emitted
    """
    tuned_list = """Available profiles:
- balanced                    - General non-specialized tuned profile
- desktop                     - Optmize for the desktop use-case
- latency-performance         - Optimize for deterministic performance
- network-latency             - Optimize for deterministic performance
- network-throughput          - Optimize for streaming network throughput.
- powersave                   - Optimize for low power-consumption
- throughput-performance      - Broadly applicable tuning that provides--
- virtual-guest               - Optimize for running inside a virtual-guest.
- virtual-host                - Optimize for running KVM guests
Current active profile: virtual-guest

** COLLECTED WARNINGS **
No SMBIOS nor DMI entry point found, sorry.
** END OF WARNINGS **
"""
    mock_cmd = MagicMock(return_value=tuned_list)
    with patch.dict(tuned.__salt__, {"cmd.run": mock_cmd}):
        assert tuned.list_() == [
            "balanced",
            "desktop",
            "latency-performance",
            "network-latency",
            "network-throughput",
            "powersave",
            "throughput-performance",
            "virtual-guest",
            "virtual-host",
        ]


def test_none():
    """ """
    ret = {
        "pid": 12345,
        "retcode": 1,
        "stderr": (
            "stderr: Cannot talk to Tuned daemon via DBus. Is Tuned daemon running?"
        ),
        "stdout": "No current active profile.",
    }
    mock_cmd = MagicMock(return_value=ret)
    with patch.dict(tuned.__salt__, {"cmd.run_all": mock_cmd}):
        assert tuned.active() == "none"



"""
Management of incron, the inotify cron
==============================================

The incron state module allows for user incrontabs to be cleanly managed.

Incron declarations require a number of parameters. The parameters needed
to be declared: ``path``, ``mask``, and ``cmd``. The ``user`` whose incrontab is to be edited
also needs to be defined.

When making changes to an existing incron job, the ``path`` declaration is the unique
factor, so if an existing cron that looks like this:

.. code-block:: yaml

    Watch for modifications in /home/user:
        incron.present:
            - user: root
            - path: /home/user
            - mask:
                - IN_MODIFY
            - cmd: 'echo "$$ $@"'

Is changed to this:

.. code-block:: yaml

    Watch for modifications and access in /home/user:
        incron.present:
            - user: root
            - path: /home/user
            - mask:
                - IN_MODIFY
                - IN_ACCESS
            - cmd: 'echo "$$ $@"'

Then the existing cron will be updated, but if the cron command is changed,
then a new cron job will be added to the user's crontab.

.. versionadded:: 0.17.0

"""

import logging

log = logging.getLogger(__name__)


def _check_cron(user, path, mask, cmd):
    """
    Return the changes
    """
    arg_mask = mask.split(",")
    arg_mask.sort()

    lst = __salt__["incron.list_tab"](user)
    if cmd.endswith("\n"):
        cmd = cmd[:-1]
    for cron in lst["crons"]:
        if path == cron["path"] and cron["cmd"] == cmd:
            cron_mask = cron["mask"].split(",")
            cron_mask.sort()
            if cron_mask == arg_mask:
                return "present"
            if any([x in cron_mask for x in arg_mask]):
                return "update"
    return "absent"


def _get_cron_info():
    """
    Returns the proper group owner and path to the incron directory
    """
    owner = "root"
    if __grains__["os"] == "FreeBSD":
        group = "wheel"
        crontab_dir = "/var/spool/incron"
    elif __grains__["os"] == "OpenBSD":
        group = "crontab"
        crontab_dir = "/var/spool/incron"
    elif __grains__.get("os_family") == "Solaris":
        group = "root"
        crontab_dir = "/var/spool/incron"
    else:
        group = "root"
        crontab_dir = "/var/spool/incron"
    return owner, group, crontab_dir


def present(name, path, mask, cmd, user="root"):
    """
    Verifies that the specified incron job is present for the specified user.
    For more advanced information about what exactly can be set in the cron
    timing parameters, check your incron system's documentation. Most Unix-like
    systems' incron documentation can be found via the incrontab man page:
    ``man 5 incrontab``.

    name
        Unique comment describing the entry

    path
        The path that should be watched

    user
        The name of the user who's crontab needs to be modified, defaults to
        the root user

    mask
        The mask of events that should be monitored for

    cmd
        The cmd that should be executed

    """
    mask = ",".join(mask)

    ret = {"changes": {}, "comment": "", "name": name, "result": True}
    if __opts__["test"]:
        status = _check_cron(user, path, mask, cmd)
        ret["result"] = None
        if status == "absent":
            ret["comment"] = f"Incron {name} is set to be added"
        elif status == "present":
            ret["result"] = True
            ret["comment"] = f"Incron {name} already present"
        elif status == "update":
            ret["comment"] = f"Incron {name} is set to be updated"
        return ret

    data = __salt__["incron.set_job"](user=user, path=path, mask=mask, cmd=cmd)
    if data == "present":
        ret["comment"] = f"Incron {name} already present"
        return ret

    if data == "new":
        ret["comment"] = f"Incron {name} added to {user}'s incrontab"
        ret["changes"] = {user: name}
        return ret

    if data == "updated":
        ret["comment"] = f"Incron {name} updated"
        ret["changes"] = {user: name}
        return ret
    ret["comment"] = "Incron {} for user {} failed to commit with error \n{}".format(
        name, user, data
    )
    ret["result"] = False
    return ret


def absent(name, path, mask, cmd, user="root"):
    """
    Verifies that the specified incron job is absent for the specified user; only
    the name is matched when removing a incron job.

    name
        Unique comment describing the entry

    path
        The path that should be watched

    user
        The name of the user who's crontab needs to be modified, defaults to
        the root user

    mask
        The mask of events that should be monitored for

    cmd
        The cmd that should be executed

    """

    mask = ",".join(mask)

    ret = {"name": name, "result": True, "changes": {}, "comment": ""}

    if __opts__["test"]:
        status = _check_cron(user, path, mask, cmd)
        ret["result"] = None
        if status == "absent":
            ret["result"] = True
            ret["comment"] = f"Incron {name} is absent"
        elif status == "present" or status == "update":
            ret["comment"] = f"Incron {name} is set to be removed"
        return ret

    data = __salt__["incron.rm_job"](user=user, path=path, mask=mask, cmd=cmd)
    if data == "absent":
        ret["comment"] = f"Incron {name} already absent"
        return ret
    if data == "removed":
        ret["comment"] = f"Incron {name} removed from {user}'s crontab"
        ret["changes"] = {user: name}
        return ret
    ret["comment"] = "Incron {} for user {} failed to commit with error {}".format(
        name, user, data
    )
    ret["result"] = False
    return ret



import pytest


@pytest.fixture(scope="package")
def salt_cloud_cli(salt_master_factory):
    """
    The ``salt-cloud`` CLI as a fixture against the running master
    """
    return salt_master_factory.salt_cloud_cli()



"""
:codeauthor: Shane Lee <slee@saltstack.com>
"""

import pytest

import salt.modules.win_file as win_file
import salt.modules.win_lgpo as win_lgpo
from tests.support.mock import MagicMock, patch

pytestmark = [
    pytest.mark.windows_whitelisted,
    pytest.mark.skip_unless_on_windows,
    pytest.mark.slow_test,
]


@pytest.fixture
def configure_loader_modules(tmp_path):
    cachedir = tmp_path / "__test_admx_policy_cache_dir"
    cachedir.mkdir(parents=True, exist_ok=True)
    return {
        win_lgpo: {
            "__salt__": {
                "file.file_exists": win_file.file_exists,
                "file.makedirs": win_file.makedirs_,
            },
            "__opts__": {
                "cachedir": str(cachedir),
            },
        },
    }


def test_get_policy_name():
    result = win_lgpo.get_policy(
        policy_name="Network firewall: Public: Settings: Display a notification",
        policy_class="machine",
        return_value_only=True,
        return_full_policy_names=True,
        hierarchical_return=False,
    )
    expected = "Not configured"
    assert result == expected


def test_get_adml_display_name_bad_name():
    result = win_lgpo._getAdmlDisplayName("junk", "spongbob")
    assert result is None


def test_get_adml_display_name_no_results():
    patch_xpath = patch.object(win_lgpo, "ADML_DISPLAY_NAME_XPATH", return_value=[])
    with patch_xpath:
        result = win_lgpo._getAdmlDisplayName("junk", "$(spongbob.squarepants)")
    assert result is None


def test_get_policy_id():
    result = win_lgpo.get_policy(
        policy_name="WfwPublicSettingsNotification",
        policy_class="machine",
        return_value_only=True,
        return_full_policy_names=True,
        hierarchical_return=False,
    )
    expected = "Not configured"
    assert result == expected


def test_get_policy_name_full_return():
    result = win_lgpo.get_policy(
        policy_name="Network firewall: Public: Settings: Display a notification",
        policy_class="machine",
        return_value_only=False,
        return_full_policy_names=True,
        hierarchical_return=False,
    )
    expected = {
        "Network firewall: Public: Settings: Display a notification": "Not configured"
    }
    assert result == expected


def test_get_policy_id_full_return():
    result = win_lgpo.get_policy(
        policy_name="WfwPublicSettingsNotification",
        policy_class="machine",
        return_value_only=False,
        return_full_policy_names=True,
        hierarchical_return=False,
    )
    expected = {
        "Network firewall: Public: Settings: Display a notification": "Not configured"
    }
    assert result == expected


def test_get_policy_name_full_return_ids():
    result = win_lgpo.get_policy(
        policy_name="Network firewall: Public: Settings: Display a notification",
        policy_class="machine",
        return_value_only=False,
        return_full_policy_names=False,
        hierarchical_return=False,
    )
    expected = {
        "Network firewall: Public: Settings: Display a notification": "Not configured"
    }
    assert result == expected


def test_get_policy_id_full_return_ids():
    result = win_lgpo.get_policy(
        policy_name="WfwPublicSettingsNotification",
        policy_class="machine",
        return_value_only=False,
        return_full_policy_names=False,
        hierarchical_return=False,
    )
    expected = {"WfwPublicSettingsNotification": "Not configured"}
    assert result == expected


def test_get_policy_id_full_return_ids_hierarchical():
    result = win_lgpo.get_policy(
        policy_name="WfwPublicSettingsNotification",
        policy_class="machine",
        return_value_only=False,
        return_full_policy_names=False,
        hierarchical_return=True,
    )
    expected = {
        "Computer Configuration": {
            "Windows Settings": {
                "Security Settings": {
                    "Windows Firewall with Advanced Security": {
                        "Windows Firewall with Advanced Security - Local Group Policy Object": {
                            "WfwPublicSettingsNotification": "Not configured"
                        }
                    }
                }
            }
        }
    }
    assert result == expected


def test_get_policy_id_full_return_full_names_hierarchical():
    result = win_lgpo.get_policy(
        policy_name="WfwPublicSettingsNotification",
        policy_class="machine",
        return_value_only=False,
        return_full_policy_names=True,
        hierarchical_return=True,
    )
    expected = {
        "Computer Configuration": {
            "Windows Settings": {
                "Security Settings": {
                    "Windows Firewall with Advanced Security": {
                        "Windows Firewall with Advanced Security - Local Group Policy Object": {
                            "Network firewall: Public: Settings: Display a notification": (
                                "Not configured"
                            )
                        }
                    }
                }
            }
        }
    }
    assert result == expected


def test_transform_value_missing_type():
    policy = {"Transform": {"some_type": "junk"}}
    result = win_lgpo._transform_value(
        value="spongebob",
        policy=policy,
        transform_type="different_type",
    )
    assert result == "spongebob"


def test_transform_value_registry():
    policy = {"Registry": {}}
    result = win_lgpo._transform_value(
        value="spongebob",
        policy=policy,
        transform_type="different_type",
    )
    assert result == "spongebob"


def test_transform_value_registry_not_set():
    policy = {"Registry": {}}
    result = win_lgpo._transform_value(
        value="(value not set)",
        policy=policy,
        transform_type="different_type",
    )
    assert result == "Not Defined"


def test_validate_setting_not_in_list():
    policy = {"Settings": ["junk"]}
    result = win_lgpo._validateSetting(value="spongebob", policy=policy)
    assert not result


def test_validate_setting_in_list():
    policy = {"Settings": ["spongebob"]}
    result = win_lgpo._validateSetting(value="spongebob", policy=policy)
    assert result


def test_validate_setting_not_list_or_dict():
    policy = {"Settings": "spongebob"}
    result = win_lgpo._validateSetting(value="spongebob", policy=policy)
    assert result


def test_add_account_rights_error():
    patch_w32sec = patch(
        "win32security.LsaOpenPolicy", MagicMock(side_effect=Exception)
    )
    with patch_w32sec:
        assert win_lgpo._addAccountRights("spongebob", "junk") is False


def test_del_account_rights_error():
    patch_w32sec = patch(
        "win32security.LsaOpenPolicy", MagicMock(side_effect=Exception)
    )
    with patch_w32sec:
        assert win_lgpo._delAccountRights("spongebob", "junk") is False


def test_validate_setting_no_function():
    policy = {
        "Settings": {
            "Function": "_in_range_inclusive",
            "Args": {"min": 0, "max": 24},
        },
    }
    result = win_lgpo._validateSetting(value="spongebob", policy=policy)
    assert not result



"""
    :codeauthor: Jayesh Kariya <jayeshk@saltstack.com>
"""

import pytest

import salt.modules.win_disk as win_disk


class MockKernel32:
    """
    Mock windll class
    """

    def __init__(self):
        pass

    @staticmethod
    def GetLogicalDrives():
        """
        Mock GetLogicalDrives method
        """
        return 1


class MockWindll:
    """
    Mock windll class
    """

    def __init__(self):
        self.kernel32 = MockKernel32()


class MockCtypes:
    """
    Mock ctypes class
    """

    def __init__(self):
        self.windll = MockWindll()


@pytest.fixture
def configure_loader_modules():
    return {win_disk: {"ctypes": MockCtypes()}}


def test_usage():
    """
    Test if it return usage information for volumes mounted on this minion.
    """
    assert win_disk.usage() == {
        "A:\\": {
            "available": None,
            "1K-blocks": None,
            "used": None,
            "capacity": None,
            "filesystem": "A:\\",
        }
    }



"""
This runner makes Salt's
execution modules available
on the salt master.

.. versionadded:: 2016.11.0

.. _salt_salt_runner:

Salt's execution modules are normally available
on the salt minion. Use this runner to call
execution modules on the salt master.
Salt :ref:`execution modules <writing-execution-modules>`
are the functions called by the ``salt`` command.

Execution modules can be called with ``salt-run``:

.. code-block:: bash

    salt-run salt.cmd test.ping
    # call functions with arguments and keyword arguments
    salt-run salt.cmd test.arg 1 2 3 key=value a=1

Execution modules are also available to salt runners:

.. code-block:: python

    __salt__['salt.cmd'](fun=fun, args=args, kwargs=kwargs)

"""

import copy
import logging

import salt.client
import salt.loader
import salt.pillar
import salt.utils.args
from salt.exceptions import SaltClientError

log = logging.getLogger(__name__)


def cmd(fun, *args, **kwargs):
    """
    .. versionchanged:: 2018.3.0
        Added ``with_pillar`` argument

    Execute ``fun`` with the given ``args`` and ``kwargs``.  Parameter ``fun``
    should be the string :ref:`name <all-salt.modules>` of the execution module
    to call.

    .. note::
        Execution modules will be loaded *every time* this function is called.
        Additionally, keep in mind that since runners execute on the master,
        custom execution modules will need to be synced to the master using
        :py:func:`salt-run saltutil.sync_modules
        <salt.runners.saltutil.sync_modules>`, otherwise they will not be
        available.

    with_pillar : False
        If ``True``, pillar data will be compiled for the master

        .. note::
            To target the master in the pillar top file, keep in mind that the
            default ``id`` for the master is ``<hostname>_master``. This can be
            overridden by setting an ``id`` configuration parameter in the
            master config file.

    CLI Example:

    .. code-block:: bash

        salt-run salt.cmd test.ping
        # call functions with arguments and keyword arguments
        salt-run salt.cmd test.arg 1 2 3 a=1
        salt-run salt.cmd mymod.myfunc with_pillar=True
    """
    log.debug("Called salt.cmd runner with minion function %s", fun)

    kwargs = salt.utils.args.clean_kwargs(**kwargs)
    with_pillar = kwargs.pop("with_pillar", False)

    opts = copy.deepcopy(__opts__)
    opts["grains"] = salt.loader.grains(opts)

    if with_pillar:
        opts["pillar"] = salt.pillar.get_pillar(
            opts,
            opts["grains"],
            opts["id"],
            saltenv=opts["saltenv"],
            pillarenv=opts.get("pillarenv"),
        ).compile_pillar()
    else:
        opts["pillar"] = {}

    functions = salt.loader.minion_mods(
        opts, utils=salt.loader.utils(opts), context=__context__
    )

    return (
        functions[fun](*args, **kwargs)
        if fun in functions
        else f"'{fun}' is not available."
    )


def execute(
    tgt,
    fun,
    arg=(),
    timeout=None,
    tgt_type="glob",
    ret="",
    jid="",
    kwarg=None,
    **kwargs,
):
    """
    .. versionadded:: 2017.7.0

    Execute ``fun`` on all minions matched by ``tgt`` and ``tgt_type``.
    Parameter ``fun`` is the name of execution module function to call.

    This function should mainly be used as a helper for runner modules,
    in order to avoid redundant code.
    For example, when inside a runner one needs to execute a certain function
    on arbitrary groups of minions, only has to:

    .. code-block:: python

        ret1 = __salt__['salt.execute']('*', 'mod.fun')
        ret2 = __salt__['salt.execute']('my_nodegroup', 'mod2.fun2', tgt_type='nodegroup')

    It can also be used to schedule jobs directly on the master, for example:

    .. code-block:: yaml

        schedule:
            collect_bgp_stats:
                function: salt.execute
                args:
                    - edge-routers
                    - bgp.neighbors
                kwargs:
                    tgt_type: nodegroup
                days: 1
                returner: redis
    """
    with salt.client.get_local_client(__opts__["conf_file"]) as client:
        try:
            return client.cmd(
                tgt,
                fun,
                arg=arg,
                timeout=timeout or __opts__["timeout"],
                tgt_type=tgt_type,  # no warn_until, as this is introduced only in 2017.7.0
                ret=ret,
                jid=jid,
                kwarg=kwarg,
                **kwargs,
            )
        except SaltClientError as client_error:
            log.error("Error while executing %s on %s (%s)", fun, tgt, tgt_type)
            log.error(client_error)
            return {}



import pytest
from saltfactories.utils import random_string

pytestmark = [
    pytest.mark.windows_whitelisted,
]


@pytest.fixture
def pillar_file_path(salt_master):
    pillar_dir = salt_master.pillar_tree.base.write_path
    testfile = pillar_dir / random_string("foo")
    try:
        yield testfile
    finally:
        if testfile.exists():
            testfile.unlink()


def test_write(client, pillar_file_path):
    ret = client.cmd(
        "pillar_roots.write", kwarg={"data": "foo: bar", "path": pillar_file_path.name}
    )
    assert pillar_file_path.is_file()
    assert ret.find("Wrote data to file") != -1


def test_write_subdir(client, salt_master):
    ret = client.cmd(
        "pillar_roots.write", kwarg={"data": "foo: bar", "path": "sub/dir/file"}
    )
    pillar_file_path = salt_master.pillar_tree.base.write_path / "sub" / "dir" / "file"
    assert pillar_file_path.is_file()
    assert ret.find("Wrote data to file") != -1


def test_cvr_2021_25282(client, pillar_file_path):
    ret = client.cmd(
        "pillar_roots.write",
        kwarg={"data": "foo", "path": f"../{pillar_file_path.name}"},
    )
    assert not pillar_file_path.parent.parent.joinpath(pillar_file_path.name).is_file()
    assert ret.find("Invalid path") != -1


def test_cvr_2021_25282_subdir(client, pillar_file_path):
    ret = client.cmd(
        "pillar_roots.write",
        kwarg={"data": "foo", "path": f"../../{pillar_file_path.name}"},
    )
    assert not pillar_file_path.parent.parent.parent.joinpath(
        pillar_file_path.name
    ).is_file()
    assert ret.find("Invalid path") != -1



"""
A simple test engine, not intended for real use but as an example
"""

import logging

import salt.utils.event
import salt.utils.json

log = logging.getLogger(__name__)


def event_bus_context(opts):
    if opts["__role"] == "master":
        event_bus = salt.utils.event.get_master_event(
            opts, opts["sock_dir"], listen=True
        )
    else:
        event_bus = salt.utils.event.get_event(
            "minion",
            opts=opts,
            sock_dir=opts["sock_dir"],
            listen=True,
        )
        log.debug("test engine started")
    return event_bus


def start():
    """
    Listen to events and write them to a log file
    """
    with event_bus_context(__opts__) as event_bus:
        while True:
            event = event_bus.get_event()
            jevent = salt.utils.json.dumps(event)
            if event:
                log.debug(jevent)



"""
tests.pytests.unit.test_version
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

Test salt's regex git describe version parsing
"""

import re

import pytest

import salt.version
from salt.version import (
    SaltStackVersion,
    SaltVersionsInfo,
    system_information,
    versions_report,
)
from tests.support.mock import MagicMock, patch

STRIP_INITIAL_NON_NUMBERS_REGEX = re.compile(r"(?:[^\d]+)?(?P<vs>.*)")


@pytest.mark.parametrize(
    "version_string,full_info, version",
    [
        ("v0.12.0-19-g767d4f9", (0, 12, 0, 0, "", 0, 19, "g767d4f9"), None),
        ("v0.12.0-85-g2880105", (0, 12, 0, 0, "", 0, 85, "g2880105"), None),
        (
            "debian/0.11.1+ds-1-3-ga0afcbd",
            (0, 11, 1, 0, "", 0, 3, "ga0afcbd"),
            "0.11.1+3.ga0afcbd",
        ),
        ("0.12.1", (0, 12, 1, 0, "", 0, 0, None), None),
        ("0.12.1", (0, 12, 1, 0, "", 0, 0, None), None),
        ("0.17.0rc1", (0, 17, 0, 0, "rc", 1, 0, None), None),
        ("v0.17.0rc1-1-g52ebdfd", (0, 17, 0, 0, "rc", 1, 1, "g52ebdfd"), None),
        ("v2014.1.4.1", (2014, 1, 4, 1, "", 0, 0, None), None),
        ("v2014.1.4.1rc3-n/a-abcdefff", (2014, 1, 4, 1, "rc", 3, -1, "abcdefff"), None),
        ("v3.4.1.1", (3, 4, 1, 1, "", 0, 0, None), None),
        ("v3000", (3000, "", 0, 0, None), "3000"),
        ("v3000.0", (3000, "", 0, 0, None), "3000"),
        ("v4518.1", (4518, 1, "", 0, 0, None), "4518.1"),
        ("v3000rc1", (3000, "rc", 1, 0, None), "3000rc1"),
        ("v3000rc1-n/a-abcdefff", (3000, "rc", 1, -1, "abcdefff"), None),
        ("3000-n/a-1e7bc8f", (3000, "", 0, -1, "1e7bc8f"), None),
        ("3000.1-n/a-1e7bc8f", (3000, 1, "", 0, -1, "1e7bc8f"), None),
        (
            "v3000nb20201214010203-1-1e7bc8f",
            (3000, "nb", 20201214010203, 1, "1e7bc8f"),
            None,
        ),
        (
            "v3000.2nb20201214010203-0-1e7bc8f",
            (3000, 2, "nb", 20201214010203, 0, "1e7bc8f"),
            "3000.2nb20201214010203",
        ),
        ("v3006.0", (3006, 0, "", 0, 0, None), "3006.0"),
        ("v3006.0rc1", (3006, 0, "rc", 1, 0, None), "3006.0rc1"),
        ("v3006.1", (3006, 1, "", 0, 0, None), "3006.1"),
        ("v3006.1rc1", (3006, 1, "rc", 1, 0, None), "3006.1rc1"),
    ],
)
def test_version_parsing(version_string, full_info, version):
    saltstack_version = SaltStackVersion.parse(version_string)
    assert saltstack_version.full_info == full_info
    if version is None:
        version = (
            STRIP_INITIAL_NON_NUMBERS_REGEX.search(version_string)
            .group("vs")  # Strip leading non numeric chars
            # Now, make it Wheel metadata 1.2 compliant post release
            .replace("n/a", "0na")
            .replace("-", "+", 1)
            .replace("-", ".", 1)
        )
    assert version in saltstack_version.string


@pytest.mark.parametrize(
    "higher_version,lower_version",
    [
        ("debian/0.11.1+ds-1-3-ga0afcbd", "0.11.1+ds-2"),
        ("v0.12.0-85-g2880105", "v0.12.0-19-g767d4f9"),
        ("v0.17.0rc1-1-g52ebdfd", "0.17.0rc1"),
        ("v0.17.0", "v0.17.0rc1"),
        ("Hydrogen", "0.17.0"),
        ("Helium", "Hydrogen"),
        ("v2014.1.4.1-n/a-abcdefff", "v2014.1.4.1rc3-n/a-abcdefff"),
        ("v2014.1.4.1-1-abcdefff", "v2014.1.4.1-n/a-abcdefff"),
        ("v2016.12.0rc1", "v2016.12.0b1"),
        ("v2016.12.0beta1", "v2016.12.0alpha1"),
        ("v2016.12.0alpha1", "v2016.12.0alpha0"),
        ("v3000.1", "v3000"),
        ("v3000rc2", "v3000rc1"),
        ("v3001", "v3000"),
        ("v4023rc1", "v4022rc1"),
        ("v3000", "v3000rc1"),
        ("v3000", "v2019.2.1"),
        ("v3000.1", "v2019.2.1"),
        # we created v3000.0rc1 tag on repo
        # but we should not be using this
        # version scheme in the future
        # but still adding test for it
        ("v3000", "v3000.0rc1"),
        ("v3000.1rc1", "v3000.0rc1"),
        ("v3000", "v2019.2.1rc1"),
        ("v3001rc1", "v2019.2.1rc1"),
        ("v3002", "v3002nb20201213"),
        ("v3002rc1", "v3002nb20201213"),
        ("v3006.0", "v3006.0rc1"),
        ("v3006.1", "v3006.0rc1"),
        ("v3006.1", "v3006.0"),
    ],
)
def test_version_comparison(higher_version, lower_version):
    assert SaltStackVersion.parse(higher_version) > lower_version
    assert SaltStackVersion.parse(lower_version) < higher_version
    assert SaltStackVersion.parse(lower_version) != higher_version


def test_unparsable_version():
    with pytest.raises(ValueError):
        SaltStackVersion.parse("Drunk")


def test_unparsable_version_from_name():
    with pytest.raises(ValueError):
        SaltStackVersion.from_name("Drunk")


@pytest.mark.parametrize(
    "commit,match",
    [
        ("d6cd1e2bd19e03a81132a23b2025920577f84e37", True),
        ("2880105", True),
        ("v3000.0.1", False),
        ("v0.12.0-85-g2880105", False),
        ("v0.12.0-0-g2880105", False),
    ],
)
def test_sha(commit, match):
    """
    test matching sha's
    """
    ret = SaltStackVersion.git_sha_regex.match(commit)
    if match:
        assert ret is not None
    else:
        assert ret is None


def test_version_report_lines():
    """
    Validate padding in versions report is correct
    """
    # Get a set of all version report name lengths including padding
    versions_report_ret = list(versions_report())
    start_looking_index = versions_report_ret.index("Dependency Versions:") + 1
    line_lengths = {
        len(line.split(":")[0])
        for line in versions_report_ret[start_looking_index:]
        if line != " " and line not in ("System Versions:", "Salt Extensions:")
    }
    # Check that they are all the same size (only one element in the set)
    assert len(line_lengths) == 1


def test_string_new_version():
    """
    Validate string property method
    using new versioning scheme
    """
    maj_ver = "3000"
    ver = SaltStackVersion(major=maj_ver)
    assert not ver.minor
    assert not ver.bugfix
    assert maj_ver == ver.string


def test_string_new_version_minor():
    """
    Validate string property method
    using new versioning scheme alongside
    minor version
    """
    maj_ver = 3000
    min_ver = 1
    ver = SaltStackVersion(major=maj_ver, minor=min_ver)
    assert ver.minor == min_ver
    assert not ver.bugfix
    assert ver.string == f"{maj_ver}.{min_ver}"


def test_string_new_version_minor_as_string():
    """
    Validate string property method
    using new versioning scheme alongside
    minor version
    """
    maj_ver = "3000"
    min_ver = "1"
    ver = SaltStackVersion(major=maj_ver, minor=min_ver)
    assert ver.minor == int(min_ver)
    assert not ver.bugfix
    assert ver.string == f"{maj_ver}.{min_ver}"

    # This only seems to happen on a cloned repo without its tags
    maj_ver = "3000"
    min_ver = ""
    ver = SaltStackVersion(major=maj_ver, minor=min_ver)
    assert ver.minor is None, f"{ver.minor!r} is not {min_ver!r}"
    assert not ver.bugfix
    assert ver.string == maj_ver


def test_string_old_version():
    """
    Validate string property method
    using old versioning scheme alongside
    minor version
    """
    maj_ver = "2019"
    min_ver = "2"
    ver = SaltStackVersion(major=maj_ver, minor=min_ver)
    assert ver.bugfix == 0
    assert ver.string == f"{maj_ver}.{min_ver}.0"


@pytest.mark.parametrize(
    "vstr,noc_info",
    [
        ("v2014.1.4.1rc3-n/a-abcdefff", (2014, 1, 4, 1, "rc", 3, -1)),
        ("v3.4.1.1", (3, 4, 1, 1, "", 0, 0)),
        ("v3000", (3000, "", 0, 0)),
        ("v3000.0", (3000, "", 0, 0)),
        ("v4518.1", (4518, 1, "", 0, 0)),
        ("v3000rc1", (3000, "rc", 1, 0)),
        ("v3000rc1-n/a-abcdefff", (3000, "rc", 1, -1)),
        ("v3000nb1-0-abcdefff", (3000, "nb", 1, 0)),
    ],
)
def test_noc_info(vstr, noc_info):
    """
    Test noc_info property method
    """
    saltstack_version = SaltStackVersion.parse(vstr)
    assert saltstack_version.noc_info, noc_info
    assert len(saltstack_version.noc_info) == len(noc_info)


@pytest.mark.parametrize(
    "vstr,full_info",
    [
        ("v2014.1.4.1rc3-n/a-abcdefff", (2014, 1, 4, 1, "rc", 3, -1, "abcdefff")),
        ("v3.4.1.1", (3, 4, 1, 1, "", 0, 0, None)),
        ("v3000", (3000, "", 0, 0, None)),
        ("v3000.0", (3000, "", 0, 0, None)),
        ("v4518.1", (4518, 1, "", 0, 0, None)),
        ("v3000rc1", (3000, "rc", 1, 0, None)),
        ("v3000rc1-n/a-abcdefff", (3000, "rc", 1, -1, "abcdefff")),
        ("v3000nb20201214-0-gabcdefff", (3000, "nb", 20201214, 0, "gabcdefff")),
    ],
)
def test_full_info(vstr, full_info):
    """
    Test full_info property method
    """
    saltstack_version = SaltStackVersion.parse(vstr)
    assert saltstack_version.full_info, full_info
    assert len(saltstack_version.full_info) == len(full_info)


@pytest.mark.parametrize(
    "vstr,full_info",
    [
        ("v2014.1.4.1rc3-n/a-abcdefff", (2014, 1, 4, 1, "rc", 3, -1, "abcdefff")),
        ("v3.4.1.1", (3, 4, 1, 1, "", 0, 0, None)),
        ("v3000", (3000, None, None, 0, "", 0, 0, None)),
        ("v3000.0", (3000, 0, None, 0, "", 0, 0, None)),
        ("v4518.1", (4518, 1, None, 0, "", 0, 0, None)),
        ("v3000rc1", (3000, None, None, 0, "rc", 2, 0, None)),
        ("v3000rc1-n/a-abcdefff", (3000, None, None, 0, "rc", 1, -1, "abcdefff")),
        ("v3000nb2-0-abcdefff", (3000, None, None, 0, "nb", 2, 0, "abcdefff")),
    ],
)
def test_full_info_all_versions(vstr, full_info):
    """
    Test full_info_all_versions property method
    """
    saltstack_version = SaltStackVersion.parse(vstr)
    assert saltstack_version.full_info_all_versions
    assert len(saltstack_version.full_info_all_versions) == len(full_info)


@pytest.mark.parametrize(
    "major,minor,tag,expected",
    [
        (3000, None, b"v3000.0rc2-12-g44fe283a77\n", "3000rc2-12-g44fe283a77"),
        (3000, None, b"v3000.0rc2-0-g44fe283a77\n", "3000rc2"),
        (3000, None, b"v3000", "3000"),
        (3000, None, b"1234567", "3000-0na-1234567"),
        (2019, 2, b"v2019.2.0rc2-12-g44fe283a77\n", "2019.2.0rc2-12-g44fe283a77"),
        (2019, 2, b"v2019.2.0", "2019.2.0"),
        (2019, 2, b"afc9830198dj", "2019.2.0-0na-afc9830198dj"),
    ],
)
def test_discover_version(major, minor, tag, expected):
    """
    Test call to __discover_version
    when using different versions
    """
    salt_ver = SaltStackVersion(major=major, minor=minor, bugfix=None)
    attrs = {
        "communicate.return_value": (tag, b""),
        "returncode.return_value": 0,
    }
    proc_ret = MagicMock(**attrs)
    proc_mock = patch("subprocess.Popen", return_value=proc_ret)
    patch_os = patch("os.path.exists", return_value=True)

    with proc_mock, patch_os:
        ret = getattr(salt.version, "__discover_version")(salt_ver)
    assert ret == expected


@pytest.mark.parametrize(
    "major,minor,bugfix",
    [
        (3000, None, None),
        (3000, 1, None),
        (3001, 0, None),
    ],
)
def test_info_new_version(major, minor, bugfix):
    """
    test info property method with new versioning scheme
    """
    ver = SaltStackVersion(major=major, minor=minor, bugfix=bugfix)
    if minor:
        assert ver.info == (major, minor)
    else:
        assert ver.info == (major,)


@pytest.mark.parametrize(
    "major,minor,bugfix",
    [
        (2019, 2, 1),
        (2018, 3, 0),
        (2017, 7, None),
    ],
)
def test_info_old_version(major, minor, bugfix):
    """
    test info property method with old versioning scheme
    """
    ver = SaltStackVersion(major=major, minor=minor, bugfix=bugfix)
    if bugfix is None:
        assert ver.info == (major, minor, 0, 0)
    else:
        assert ver.info == (major, minor, bugfix, 0)


def test_bugfix_string():
    """
    test when bugfix is an empty string
    """
    ret = SaltStackVersion(3000, 1, "", 0, 0, None)
    assert ret.info == (3000, 1)
    assert ret.minor == 1
    assert ret.bugfix is None


@pytest.mark.parametrize(
    "version_tuple,expected",
    [
        (
            (3000, 1, None, None, "", 0, 0, None),
            "<SaltStackVersion name='Neon' major=3000 minor=1>",
        ),
        (
            (3000, 0, None, None, "", 0, 0, None),
            "<SaltStackVersion name='Neon' major=3000>",
        ),
        (
            (2019, 2, 3, None, "", 0, 0, None),
            "<SaltStackVersion name='Fluorine' major=2019 minor=2 bugfix=3>",
        ),
        (
            (2019, 2, 3, None, "rc", 1, 0, None),
            "<SaltStackVersion name='Fluorine' major=2019 minor=2 bugfix=3 rc=1>",
        ),
        (
            (2019, 2, 3, None, "nb", 20201214, 0, None),
            "<SaltStackVersion name='Fluorine' major=2019 minor=2 bugfix=3"
            " nb=20201214>",
        ),
    ],
)
def test_version_repr(version_tuple, expected):
    """
    Test SaltStackVersion repr for both date
    and new versioning scheme
    """
    assert repr(SaltStackVersion(*version_tuple)) == expected


def test_previous_and_next_releases():
    with patch.multiple(
        SaltVersionsInfo,
        _previous_release=None,
        _next_release=None,
        _current_release=SaltVersionsInfo.CALIFORNIUM,
    ):
        assert SaltVersionsInfo.current_release() == SaltVersionsInfo.CALIFORNIUM
        assert SaltVersionsInfo.next_release() == SaltVersionsInfo.EINSTEINIUM
        assert SaltVersionsInfo.previous_release() == SaltVersionsInfo.BERKELIUM

    with patch.multiple(
        SaltVersionsInfo,
        _previous_release=None,
        _next_release=None,
        _current_release=SaltVersionsInfo.NEPTUNIUM,
    ):
        assert SaltVersionsInfo.current_release() == SaltVersionsInfo.NEPTUNIUM
        assert SaltVersionsInfo.next_release() == SaltVersionsInfo.PLUTONIUM
        assert SaltVersionsInfo.previous_release() == SaltVersionsInfo.URANIUM


@pytest.mark.skip_unless_on_linux
def test_system_version_linux():
    """
    version.system_version on Linux
    """

    with patch(
        "salt.utils.platform.linux_distribution",
        MagicMock(return_value=("Manjaro Linux", "20.0.2", "Lysia")),
    ):
        versions = [item for item in system_information()]
        version = ("version", "Manjaro Linux 20.0.2 Lysia")
        assert version in versions

    with patch(
        "salt.utils.platform.linux_distribution",
        MagicMock(return_value=("Debian GNU/Linux", "9", "stretch")),
    ):
        versions = [item for item in system_information()]
        version = ("version", "Debian GNU/Linux 9 stretch")
        assert version in versions

    with patch(
        "salt.utils.platform.linux_distribution",
        MagicMock(return_value=("Debian GNU/Linux", "10", "buster")),
    ):
        versions = [item for item in system_information()]
        version = ("version", "Debian GNU/Linux 10 buster")
        assert version in versions

    with patch(
        "salt.utils.platform.linux_distribution",
        MagicMock(return_value=("CentOS Linux", "7", "Core")),
    ):
        versions = [item for item in system_information()]
        version = ("version", "CentOS Linux 7 Core")
        assert version in versions

    with patch(
        "salt.utils.platform.linux_distribution",
        MagicMock(return_value=("CentOS Linux", "8", "Core")),
    ):
        versions = [item for item in system_information()]
        version = ("version", "CentOS Linux 8 Core")
        assert version in versions

    with patch(
        "salt.utils.platform.linux_distribution",
        MagicMock(return_value=("OpenSUSE Leap", "15.1", "")),
    ):
        versions = [item for item in system_information()]
        version = ("version", "OpenSUSE Leap 15.1 ")
        assert version in versions


@pytest.mark.skip_unless_on_darwin
def test_system_version_osx():
    """
    version.system_version on OS X
    """

    with patch(
        "platform.mac_ver",
        MagicMock(return_value=("10.15.2", ("", "", ""), "x86_64")),
    ):
        versions = [item for item in system_information()]
        version = ("version", "10.15.2 x86_64")
        assert version in versions


@pytest.mark.skip_unless_on_windows
def test_system_version_windows():
    """
    version.system_version on Windows
    """

    with patch(
        "platform.win32_ver",
        return_value=("10", "10.0.14393", "SP0", "Multiprocessor Free"),
    ), patch("win32api.RegOpenKey", MagicMock()), patch(
        "win32api.RegQueryValueEx",
        MagicMock(return_value=("Windows Server 2016 Datacenter", 1)),
    ):
        versions = [item for item in system_information()]
        version = ("version", "2016Server 10.0.14393 SP0 Multiprocessor Free")
        assert version in versions


def test_versions_report_includes_salt_extensions():
    with patch(
        "salt.version.extensions_information", return_value={"foo-bar-ext": "1.0"}
    ):
        versions_information = salt.version.versions_information()
        assert "Salt Extensions" in versions_information
        assert "foo-bar-ext" in versions_information["Salt Extensions"]
        assert versions_information["Salt Extensions"]["foo-bar-ext"] == "1.0"


def test_versions_report_no_extensions_available():
    with patch("salt.utils.entrypoints.iter_entry_points", return_value=()):
        versions_information = salt.version.versions_information()
        assert "Salt Extensions" not in versions_information


@pytest.mark.parametrize(
    "version_str,expected_str,expected_name",
    [
        ("2014.1.4", "2014.1.4", "Hydrogen"),
        ("3000.1", "3000.1", "Neon"),
        ("3005", "3005", "Phosphorus"),
        ("3006", "3006.0", "Sulfur"),
        ("3006.0", "3006.0", "Sulfur"),
        ("3006.1", "3006.1", "Sulfur"),
        ("3015.1", "3015.1", "Manganese"),
        ("3109.3", "3109.3", None),
    ],
)
def test_parsed_version_name(version_str, expected_str, expected_name):
    """
    Test all versioning schemes name attribute.

    The old, new, and dot zero, must properly set the version name attribut
    test info property method with new versioning scheme
    """
    ver = SaltStackVersion.parse(version_str)
    assert str(ver) == expected_str
    if expected_name:
        assert ver.name == expected_name
    else:
        assert ver.name is None



"""
    :codeauthor: Jayesh Kariya <jayeshk@saltstack.com>
"""

import pytest

import salt.states.portage_config as portage_config
from tests.support.mock import MagicMock, patch


@pytest.fixture
def configure_loader_modules():
    return {portage_config: {"__opts__": {"test": True}}}


def test_mod_init():
    """
    Test to enforce a nice structure on the configuration files.
    """
    name = "salt"

    mock = MagicMock(side_effect=[True, Exception])
    with patch.dict(
        portage_config.__salt__, {"portage_config.enforce_nice_config": mock}
    ):
        assert portage_config.mod_init(name)

        assert not portage_config.mod_init(name)


def test_flags():
    """
    Test to enforce the given flags on the given package or ``DEPEND`` atom.
    """
    with patch("traceback.format_exc", MagicMock(return_value="SALTSTACK")):
        name = "salt"

        ret = {"name": name, "result": False, "comment": "SALTSTACK", "changes": {}}

        mock = MagicMock(side_effect=Exception("error"))
        with patch.dict(
            portage_config.__salt__, {"portage_config.get_missing_flags": mock}
        ):
            assert portage_config.flags(name, use="openssl") == ret

            assert portage_config.flags(name, accept_keywords=True) == ret

            assert portage_config.flags(name, env=True) == ret

            assert portage_config.flags(name, license=True) == ret

            assert portage_config.flags(name, properties=True) == ret

            assert portage_config.flags(name, mask=True) == ret

            assert portage_config.flags(name, unmask=True) == ret

            ret.update({"comment": "", "result": True})
            assert portage_config.flags(name) == ret



"""
noop_returner
~~~~~~~~~~~~~

A returner that does nothing which is used to test the salt-master `event_return` functionality
"""

import logging

import salt.utils.jid

log = logging.getLogger(__name__)

__virtualname__ = "runtests_noop"


def __virtual__():
    return True


def event_return(events):
    log.debug("NOOP_RETURN.event_return - Events: %s", events)


def returner(ret):
    log.debug("NOOP_RETURN.returner - Ret: %s", ret)


def prep_jid(nocache=False, passed_jid=None):  # pylint: disable=unused-argument
    """
    Do any work necessary to prepare a JID, including sending a custom id
    """
    return passed_jid if passed_jid is not None else salt.utils.jid.gen_jid(__opts__)



import pathlib

import pytest

from tests.support.runtests import RUNTIME_VARS

pytestmark = [
    pytest.mark.slow_test,
]


@pytest.fixture(scope="module")
def reset_pillar(salt_call_cli):
    try:
        # Run tests
        yield
    finally:
        # Refresh pillar once all tests are done.
        ret = salt_call_cli.run("saltutil.refresh_pillar", wait=True)
        assert ret.returncode == 0
        assert ret.data is True


@pytest.fixture(scope="module")
def pillar_test_true(
    base_env_pillar_tree_root_dir, salt_minion, salt_call_cli, reset_pillar
):
    top_file = """
    base:
      '{}':
        - basic
    """.format(
        salt_minion.id
    )
    basic_pillar_file = """
    test: true
    """
    with pytest.helpers.temp_file(
        "top.sls", top_file, base_env_pillar_tree_root_dir
    ), pytest.helpers.temp_file(
        "basic.sls", basic_pillar_file, base_env_pillar_tree_root_dir
    ):
        ret = salt_call_cli.run("saltutil.refresh_pillar", wait=True)
        assert ret.returncode == 0
        assert ret.data is True
        yield


@pytest.fixture(scope="module")
def pillar_test_empty(
    base_env_pillar_tree_root_dir, salt_minion, salt_call_cli, reset_pillar
):
    top_file = """
    base:
      '{}':
        - basic
    """.format(
        salt_minion.id
    )
    basic_pillar_file = ""
    with pytest.helpers.temp_file(
        "top.sls", top_file, base_env_pillar_tree_root_dir
    ), pytest.helpers.temp_file(
        "basic.sls", basic_pillar_file, base_env_pillar_tree_root_dir
    ):
        ret = salt_call_cli.run("saltutil.refresh_pillar", wait=True)
        assert ret.returncode == 0
        assert ret.data is True
        yield


@pytest.fixture(scope="module")
def pillar_test_false(
    base_env_pillar_tree_root_dir, salt_minion, salt_call_cli, reset_pillar
):
    top_file = """
    base:
      '{}':
        - basic
    """.format(
        salt_minion.id
    )
    basic_pillar_file = """
    test: false
    """
    with pytest.helpers.temp_file(
        "top.sls", top_file, base_env_pillar_tree_root_dir
    ), pytest.helpers.temp_file(
        "basic.sls", basic_pillar_file, base_env_pillar_tree_root_dir
    ):
        ret = salt_call_cli.run("saltutil.refresh_pillar", wait=True)
        assert ret.returncode == 0
        assert ret.data is True
        yield


@pytest.fixture
def testfile_path(tmp_path, base_env_state_tree_root_dir):
    testfile = tmp_path / "testfile"
    sls_contents = """
    {}:
      file:
        - managed
        - source: salt://testfile
        - makedirs: true
    """.format(
        testfile
    )
    with pytest.helpers.temp_file(
        "sls-id-test.sls", sls_contents, base_env_state_tree_root_dir
    ):
        yield testfile


@pytest.mark.usefixtures("pillar_test_true")
def test_state_sls_id_test(salt_call_cli, testfile_path):
    """
    test state.sls_id when test is set
    to true in pillar data
    """
    expected_comment = (
        "The file {} is set to be changed\nNote: No changes made, actual changes may\n"
        "be different due to other states."
    ).format(testfile_path)
    ret = salt_call_cli.run("state.sls", "sls-id-test")
    assert ret.returncode == 0
    for val in ret.data.values():
        assert val["comment"] == expected_comment
        assert val["changes"] == {"newfile": str(testfile_path)}


@pytest.mark.usefixtures("pillar_test_true")
def test_state_sls_id_test_state_test_post_run(salt_call_cli, testfile_path):
    """
    test state.sls_id when test is set to
    true post the state already being run previously
    """
    source = pathlib.Path(RUNTIME_VARS.BASE_FILES, "testfile")
    testfile_path.write_text(source.read_text(encoding="utf-8"), encoding="utf-8")
    testfile_path.chmod(0o644)
    ret = salt_call_cli.run("state.sls", "sls-id-test")
    assert ret.returncode == 0
    for val in ret.data.values():
        assert val["comment"] == "The file {} is in the correct state".format(
            testfile_path
        )
        assert val["changes"] == {}


@pytest.mark.usefixtures("pillar_test_empty")
def test_state_sls_id_test_true(salt_call_cli, testfile_path):
    """
    test state.sls_id when test=True is passed as arg
    """
    expected_comment = (
        "The file {} is set to be changed\nNote: No changes made, actual changes may\n"
        "be different due to other states."
    ).format(testfile_path)
    ret = salt_call_cli.run("state.sls", "sls-id-test", test=True)
    assert ret.returncode == 0
    for val in ret.data.values():
        assert val["comment"] == expected_comment
        assert val["changes"] == {"newfile": str(testfile_path)}


@pytest.mark.usefixtures("pillar_test_empty")
def test_state_sls_id_test_true_post_run(salt_call_cli, testfile_path):
    """
    test state.sls_id when test is set to true as an
    arg post the state already being run previously
    """
    ret = salt_call_cli.run("state.sls", "sls-id-test")
    assert ret.returncode == 0
    assert testfile_path.exists()
    for val in ret.data.values():
        assert val["comment"] == f"File {testfile_path} updated"
        assert val["changes"]["diff"] == "New file"

    ret = salt_call_cli.run("state.sls", "sls-id-test", test=True)
    assert ret.returncode == 0
    for val in ret.data.values():
        assert val["comment"] == "The file {} is in the correct state".format(
            testfile_path
        )
        assert val["changes"] == {}


@pytest.mark.usefixtures("pillar_test_true")
def test_state_sls_id_test_false_pillar_true(salt_call_cli, testfile_path):
    """
    test state.sls_id when test is set to false as an
    arg and minion_state_test is set to True. Should
    return test=False.
    """
    ret = salt_call_cli.run("state.sls", "sls-id-test", test=False)
    assert ret.returncode == 0
    for val in ret.data.values():
        assert val["comment"] == f"File {testfile_path} updated"
        assert val["changes"]["diff"] == "New file"


@pytest.mark.usefixtures("pillar_test_false")
def test_state_test_pillar_false(salt_call_cli, testfile_path):
    """
    test state.test forces test kwarg to True even when pillar is set to False
    """
    expected_comment = (
        "The file {} is set to be changed\nNote: No changes made, actual changes may\n"
        "be different due to other states."
    ).format(testfile_path)
    ret = salt_call_cli.run("state.test", "sls-id-test")
    assert ret.returncode == 0
    for val in ret.data.values():
        assert val["comment"] == expected_comment
        assert val["changes"] == {"newfile": str(testfile_path)}


@pytest.mark.usefixtures("pillar_test_false")
def test_state_test_test_false_pillar_false(salt_call_cli, testfile_path):
    """
    test state.test forces test kwarg to True even when pillar and kwarg are set
    to False
    """
    expected_comment = (
        "The file {} is set to be changed\nNote: No changes made, actual changes may\n"
        "be different due to other states."
    ).format(testfile_path)
    ret = salt_call_cli.run("state.test", "sls-id-test", test=False)
    assert ret.returncode == 0
    for val in ret.data.values():
        assert val["comment"] == expected_comment
        assert val["changes"] == {"newfile": str(testfile_path)}



import logging

import salt.utils.files
from salt.channel.client import ReqChannel

log = logging.getLogger(__name__)


class LocalChannel(ReqChannel):
    """
    Local channel for testing purposes
    """

    def __init__(self, opts, **kwargs):
        self.opts = opts
        self.kwargs = kwargs
        self.tries = 0

    def close(self):
        """
        Close the local channel.

        Currently a NOOP
        """

    def send(self, load, tries=3, timeout=60, raw=False):

        if self.tries == 0:
            log.debug("LocalChannel load: %s", load)
            # data = json.loads(load)
            # {'path': 'apt-cacher-ng/map.jinja', 'saltenv': 'base', 'cmd': '_serve_file', 'loc': 0}
            # f = open(data['path'])
            with salt.utils.files.fopen(load["path"]) as f:
                ret = {
                    "data": "".join(f.readlines()),
                    "dest": load["path"],
                }
                print("returning", ret)
        else:
            # end of buffer
            ret = {
                "data": None,
                "dest": None,
            }
        self.tries = self.tries + 1
        return ret

    def crypted_transfer_decode_dictentry(
        self, load, dictkey=None, tries=3, timeout=60
    ):
        super().crypted_transfer_decode_dictentry(
            load, dictkey=dictkey, tries=tries, timeout=timeout
        )



import pytest

import salt.utils.event
from salt.netapi.rest_tornado import saltnado
from tests.support.events import eventpublisher_process


def _check_skip(grains):
    if grains["os"] == "MacOS":
        return True
    return False


pytestmark = [
    pytest.mark.slow_test,
    pytest.mark.skip_initial_gh_actions_failure(skip=_check_skip),
]


class Request:
    __slots__ = ("_finished",)

    def __init__(self):
        self._finished = True


@pytest.fixture
def sock_dir(tmp_path):
    yield str(tmp_path)


async def test_simple(sock_dir):
    """
    Test getting a few events
    """
    with eventpublisher_process(sock_dir):
        with salt.utils.event.MasterEvent(sock_dir) as me:
            request = Request()
            event_listener = saltnado.EventListener(
                {},  # we don't use mod_opts, don't save?
                {"sock_dir": sock_dir, "transport": "zeromq"},
            )
            event_future = event_listener.get_event(
                request, "evt1"
            )  # get an event future
            me.fire_event({"data": "foo2"}, "evt2")  # fire an event we don't want
            me.fire_event({"data": "foo1"}, "evt1")  # fire an event we do want

            await event_future  # wait for the future

            # check that we got the event we wanted
            assert event_future.done()
            assert event_future.result()["tag"] == "evt1"
            assert event_future.result()["data"]["data"] == "foo1"


async def test_set_event_handler(sock_dir):
    """
    Test subscribing events using set_event_handler
    """
    with eventpublisher_process(sock_dir):
        with salt.utils.event.MasterEvent(sock_dir) as me:
            request = Request()
            event_listener = saltnado.EventListener(
                {},  # we don't use mod_opts, don't save?
                {"sock_dir": sock_dir, "transport": "zeromq"},
            )
            event_future = event_listener.get_event(
                request,
                tag="evt",
                timeout=1,
            )  # get an event future
            me.fire_event({"data": "foo"}, "evt")  # fire an event we do want

            await event_future  # wait for the future

            # check that we subscribed the event we wanted
            assert len(event_listener.timeout_map) == 0


async def test_timeout(sock_dir):
    """
    Make sure timeouts work correctly
    """
    with eventpublisher_process(sock_dir):
        request = Request()
        event_listener = saltnado.EventListener(
            {},  # we don't use mod_opts, don't save?
            {"sock_dir": sock_dir, "transport": "zeromq"},
        )
        event_future = event_listener.get_event(
            request,
            tag="evt1",
            timeout=1,
        )  # get an event future

        with pytest.raises(saltnado.TimeoutException):
            await event_future  # wait for the future

        assert event_future.done()


async def test_clean_by_request(sock_dir, io_loop):
    """
    Make sure the method clean_by_request clean up every related data in EventListener
    request_future_1 : will be timeout-ed by clean_by_request(request1)
    request_future_2 : will be finished by me.fire_event and awaiting for it ...
    request_future_3 : will be finished by me.fire_event and awaiting for it ...
    request_future_4 : will be timeout-ed by clean-by_request(request2)
    """

    with eventpublisher_process(sock_dir):
        with salt.utils.event.MasterEvent(sock_dir) as me:
            request1 = Request()
            request2 = Request()
            event_listener = saltnado.EventListener(
                {},  # we don't use mod_opts, don't save?
                {"sock_dir": sock_dir, "transport": "zeromq"},
            )

            assert 0 == len(event_listener.tag_map)
            assert 0 == len(event_listener.request_map)

            request_future_1 = event_listener.get_event(request1, tag="evt1")
            request_future_2 = event_listener.get_event(request1, tag="evt2")
            dummy_request_future_1 = event_listener.get_event(request2, tag="evt3")
            dummy_request_future_2 = event_listener.get_event(
                request2, timeout=10, tag="evt4"
            )

            assert 4 == len(event_listener.tag_map)
            assert 2 == len(event_listener.request_map)

            me.fire_event({"data": "foo2"}, "evt2")
            me.fire_event({"data": "foo3"}, "evt3")

            await request_future_2
            await dummy_request_future_1

            event_listener.clean_by_request(request1)
            me.fire_event({"data": "foo1"}, "evt1")

            assert request_future_1.done()
            with pytest.raises(saltnado.TimeoutException):
                request_future_1.result()

            assert request_future_2.done()
            assert request_future_2.result()["tag"] == "evt2"
            assert request_future_2.result()["data"]["data"] == "foo2"

            assert dummy_request_future_1.done()
            assert dummy_request_future_1.result()["tag"] == "evt3"
            assert dummy_request_future_1.result()["data"]["data"] == "foo3"

            assert not dummy_request_future_2.done()

            assert 2 == len(event_listener.tag_map)
            assert 1 == len(event_listener.request_map)

            event_listener.clean_by_request(request2)

            with pytest.raises(saltnado.TimeoutException):
                dummy_request_future_2.result()

            assert 0 == len(event_listener.tag_map)
            assert 0 == len(event_listener.request_map)



"""
Hetzner Cloud Module
====================

The Hetzner cloud module is used to control access to the hetzner cloud.
https://docs.hetzner.cloud/

:depends: hcloud >= 1.10

Use of this module requires the ``key`` parameter to be set.

.. code-block:: yaml

    my-hetzner-cloud-config:
      key: <your api key>
      driver: hetzner

"""

# pylint: disable=invalid-name,function-redefined


import logging
import time

import salt.config as config
from salt.exceptions import SaltCloudException, SaltCloudSystemExit

# hcloud module will be needed
# pylint: disable=import-error
try:
    import hcloud

    HAS_HCLOUD = True
except ImportError:
    HAS_HCLOUD = False


# Get logging started
log = logging.getLogger(__name__)

__virtualname__ = "hetzner"


def __virtual__():
    """
    Check for hetzner configurations
    """
    if get_configured_provider() is False:
        return False

    if get_dependencies() is False:
        return False

    return __virtualname__


def _get_active_provider_name():
    try:
        return __active_provider_name__.value()
    except AttributeError:
        return __active_provider_name__


def get_configured_provider():
    """
    Return the first configured instance.
    """
    return config.is_provider_configured(
        __opts__,
        _get_active_provider_name() or __virtualname__,
        ("key",),
    )


def get_dependencies():
    """
    Warn if dependencies aren't met.
    """
    return config.check_driver_dependencies(
        _get_active_provider_name() or __virtualname__,
        {"hcloud": HAS_HCLOUD},
    )


def _object_to_dict(obj, attrs):
    return {attr: getattr(obj, attr) for attr in attrs}


def _datacenter_to_dict(datacenter):
    return {
        "name": datacenter.name,
        "location": datacenter.location.name,
    }


def _public_network_to_dict(net):
    return {
        "ipv4": getattr(net.ipv4, "ip", None),
        "ipv6": getattr(net.ipv6, "ip", None),
    }


def _private_network_to_dict(net):
    return {
        "ip": getattr(net, "ip", None),
    }


def _connect_client():
    provider = get_configured_provider()
    return hcloud.Client(provider["key"])


def avail_locations(call=None):
    """
    Return a dictionary of available locations
    """
    if call == "action":
        raise SaltCloudSystemExit(
            "The list_locations function must be called with -f or --function"
        )

    client = _connect_client()
    locations = {}
    for loc in client.locations.get_all():
        locations[loc.name] = _object_to_dict(loc, loc.model.__slots__)
    return locations


def avail_images(call=None):
    """
    Return a dictionary of available images
    """
    if call == "action":
        raise SaltCloudSystemExit(
            "The avail_images function must be called with -f or --function"
        )

    client = _connect_client()
    images = {}
    for image in client.images.get_all():
        images[image.name] = _object_to_dict(image, image.model.__slots__)
    return images


def avail_sizes(call=None):
    """
    Return a dictionary of available VM sizes
    """
    if call == "action":
        raise SaltCloudSystemExit(
            "The avail_sizes function must be called with -f or --function"
        )

    client = _connect_client()
    sizes = {}
    for size in client.server_types.get_all():
        sizes[size.name] = _object_to_dict(size, size.model.__slots__)
    return sizes


def list_ssh_keys(call=None):
    """
    Return a dictionary of available SSH keys configured in the current project
    """
    if call == "action":
        raise SaltCloudSystemExit(
            "The list_ssh_keys function must be called with -f or --function"
        )

    client = _connect_client()
    ssh_keys = {}
    for key in client.ssh_keys.get_all():
        ssh_keys[key.name] = _object_to_dict(key, key.model.__slots__)
    return ssh_keys


def list_nodes_full(call=None):
    """
    Return a dictionary of existing VMs in the current project, containing full details per VM
    """
    if call == "action":
        raise SaltCloudSystemExit(
            "The list_nodes_full function must be called with -f or --function"
        )

    client = _connect_client()
    nodes = {}
    for node in client.servers.get_all():
        nodes[node.name] = {
            "id": node.id,
            "name": node.name,
            "image": node.image.name,
            "size": node.server_type.name,
            "state": node.status,
            "public_ips": _public_network_to_dict(node.public_net),
            "private_ips": list(map(_private_network_to_dict, node.private_net)),
            "labels": node.labels,
            "created": str(node.created),
            "datacenter": _datacenter_to_dict(node.datacenter),
            "volumes": [vol.name for vol in node.volumes],
        }
    return nodes


def list_nodes(call=None):
    """
    Return a dictionary of existing VMs in the current project, containing basic details of each VM
    """
    if call == "action":
        raise SaltCloudSystemExit(
            "The list_nodes function must be called with -f or --function"
        )

    ret = {}

    nodes = list_nodes_full()
    for node in nodes:
        ret[node] = {"name": node}
        for prop in ("id", "image", "size", "state", "private_ips", "public_ips"):
            ret[node][prop] = nodes[node].get(prop)
    return ret


def wait_until(name, state, timeout=300):
    """
    Wait until a specific state has been reached on  a node
    """
    start_time = time.time()
    node = show_instance(name, call="action")
    while True:
        if node["state"] == state:
            return True
        time.sleep(1)
        if time.time() - start_time > timeout:
            return False
        node = show_instance(name, call="action")


def show_instance(name, call=None):
    """
    Return the details of a specific VM
    """
    if call != "action":
        raise SaltCloudSystemExit(
            "The show_instance function must be called with -a or --action."
        )

    try:
        node = list_nodes_full("function")[name]
    except KeyError:
        log.debug("Failed to get data for node '%s'", name)
        node = {}

    __utils__["cloud.cache_node"](
        node,
        _get_active_provider_name() or __virtualname__,
        __opts__,
    )

    return node


def create(vm_):
    """
    Create a single VM from a data dict
    """
    try:
        # Check for required profile parameters before sending any API calls.
        if (
            vm_.get("profile")
            and config.is_profile_configured(
                __opts__,
                _get_active_provider_name() or __virtualname__,
                vm_["profile"],
                vm_=vm_,
            )
            is False
        ):
            return False
    except AttributeError:
        pass

    client = _connect_client()

    name = config.get_cloud_config_value(
        "name",
        vm_,
        __opts__,
        search_global=False,
    )
    if not name:
        raise SaltCloudException("Missing server name")

    # Get the required configuration
    server_type = client.server_types.get_by_name(
        config.get_cloud_config_value(
            "size",
            vm_,
            __opts__,
            search_global=False,
        )
    )
    if server_type is None:
        raise SaltCloudException("The server size is not supported")

    image = client.images.get_by_name(
        config.get_cloud_config_value(
            "image",
            vm_,
            __opts__,
            search_global=False,
        )
    )
    if image is None:
        raise SaltCloudException("The server image is not supported")

    __utils__["cloud.fire_event"](
        "event",
        "starting create",
        "salt/cloud/{}/creating".format(vm_["name"]),
        args=__utils__["cloud.filter_event"](
            "creating",
            vm_,
            ["name", "profile", "provider", "driver"],
        ),
        sock_dir=__opts__["sock_dir"],
        transport=__opts__["transport"],
    )

    # Get the ssh_keys
    ssh_keys = config.get_cloud_config_value(
        "ssh_keys", vm_, __opts__, search_global=False
    )

    if ssh_keys:
        names, ssh_keys = ssh_keys[:], []
        for n in names:
            ssh_key = client.ssh_keys.get_by_name(n)
            if ssh_key is None:
                log.error("Invalid ssh key %s.", n)
            else:
                ssh_keys.append(ssh_key)

    # Get the location
    location = config.get_cloud_config_value(
        "location",
        vm_,
        __opts__,
        search_global=False,
    )
    if location:
        location = client.locations.get_by_name(location)

        if location is None:
            raise SaltCloudException("The server location is not supported")

    # Get the datacenter
    datacenter = config.get_cloud_config_value(
        "datacenter",
        vm_,
        __opts__,
        search_global=False,
    )
    if datacenter:
        datacenter = client.datacenters.get_by_name(datacenter)

        if datacenter is None:
            raise SaltCloudException("The server datacenter is not supported")

    # Get the volumes
    volumes = config.get_cloud_config_value(
        "volumes",
        vm_,
        __opts__,
        search_global=False,
    )
    if volumes:
        volumes = [vol for vol in client.volumes.get_all() if vol in volumes]

    # Get the networks
    networks = config.get_cloud_config_value(
        "networks",
        vm_,
        __opts__,
        search_global=False,
    )
    if networks:
        networks = [vol for vol in client.networks.get_all() if vol in networks]

    # Create the machine
    response = client.servers.create(
        name=name,
        server_type=server_type,
        image=image,
        ssh_keys=ssh_keys,
        volumes=volumes,
        networks=networks,
        location=location,
        datacenter=datacenter,
        user_data=config.get_cloud_config_value(
            "user_data",
            vm_,
            __opts__,
            search_global=False,
        ),
        labels=config.get_cloud_config_value(
            "labels",
            vm_,
            __opts__,
            search_global=False,
        ),
        automount=config.get_cloud_config_value(
            "automount",
            vm_,
            __opts__,
            search_global=False,
        ),
    )

    # Bootstrap if ssh keys are configured
    server = response.server
    vm_.update(
        {
            "ssh_host": server.public_net.ipv4.ip or server.public_net.ipv6.ip,
            "ssh_password": response.root_password,
            "key_filename": config.get_cloud_config_value(
                "private_key", vm_, __opts__, search_global=False, default=None
            ),
        }
    )

    ret = __utils__["cloud.bootstrap"](vm_, __opts__)

    log.info("Created Cloud VM '%s'", vm_["name"])
    ret["created"] = True

    __utils__["cloud.fire_event"](
        "event",
        "created instance",
        "salt/cloud/{}/created".format(vm_["name"]),
        args=__utils__["cloud.filter_event"](
            "created",
            vm_,
            ["name", "profile", "provider", "driver"],
        ),
        sock_dir=__opts__["sock_dir"],
        transport=__opts__["transport"],
    )

    return ret


def start(name, call=None, wait=True):
    """
    Start a node.

    CLI Example:

    .. code-block:: bash

        salt-cloud -a start mymachine
    """
    if call != "action":
        raise SaltCloudSystemExit(
            "The start action must be called with -a or --action."
        )

    client = _connect_client()
    server = client.servers.get_by_name(name)
    if server is None:
        return f"Instance {name} doesn't exist."

    server.power_on()
    if wait and not wait_until(name, "running"):
        return f"Instance {name} doesn't start."

    __utils__["cloud.fire_event"](
        "event",
        "started instance",
        f"salt/cloud/{name}/started",
        args={"name": name},
        sock_dir=__opts__["sock_dir"],
        transport=__opts__["transport"],
    )

    return {"Started": f"{name} was started."}


def stop(name, call=None, wait=True):
    """
    Stop a node.

    CLI Example:

    .. code-block:: bash

        salt-cloud -a stop mymachine
    """
    if call != "action":
        raise SaltCloudSystemExit("The stop action must be called with -a or --action.")

    client = _connect_client()
    server = client.servers.get_by_name(name)
    if server is None:
        return f"Instance {name} doesn't exist."

    server.power_off()
    if wait and not wait_until(name, "off"):
        return f"Instance {name} doesn't stop."

    __utils__["cloud.fire_event"](
        "event",
        "stopped instance",
        f"salt/cloud/{name}/stopped",
        args={"name": name},
        sock_dir=__opts__["sock_dir"],
        transport=__opts__["transport"],
    )

    return {"Stopped": f"{name} was stopped."}


def reboot(name, call=None, wait=True):
    """
    Reboot a node.

    CLI Example:

    .. code-block:: bash

        salt-cloud -a reboot mymachine
    """
    if call != "action":
        raise SaltCloudSystemExit(
            "The reboot action must be called with -a or --action."
        )

    client = _connect_client()
    server = client.servers.get_by_name(name)
    if server is None:
        return f"Instance {name} doesn't exist."

    server.reboot()

    if wait and not wait_until(name, "running"):
        return f"Instance {name} doesn't start."

    return {"Rebooted": f"{name} was rebooted."}


def destroy(name, call=None):
    """
    Destroy a node.

    CLI Example:

    .. code-block:: bash

        salt-cloud --destroy mymachine
    """
    if call == "function":
        raise SaltCloudSystemExit(
            "The destroy action must be called with -d, --destroy, -a or --action."
        )

    client = _connect_client()
    server = client.servers.get_by_name(name)
    if server is None:
        return f"Instance {name} doesn't exist."

    __utils__["cloud.fire_event"](
        "event",
        "destroying instance",
        f"salt/cloud/{name}/destroying",
        args={"name": name},
        sock_dir=__opts__["sock_dir"],
        transport=__opts__["transport"],
    )

    node = show_instance(name, call="action")
    if node["state"] == "running":
        stop(name, call="action", wait=False)
        if not wait_until(name, "off"):
            return {"Error": f"Unable to destroy {name}, command timed out"}

    server.delete()

    __utils__["cloud.fire_event"](
        "event",
        "destroyed instance",
        f"salt/cloud/{name}/destroyed",
        args={"name": name},
        sock_dir=__opts__["sock_dir"],
        transport=__opts__["transport"],
    )

    if __opts__.get("update_cachedir", False) is True:
        __utils__["cloud.delete_minion_cachedir"](
            name,
            _get_active_provider_name().split(":")[0],
            __opts__,
        )

    return {"Destroyed": f"{name} was destroyed."}


def resize(name, kwargs, call=None):
    """
    Resize a node.

    CLI Example:

    .. code-block:: bash

        salt-cloud -a resize mymachine size=...
    """
    if call != "action":
        raise SaltCloudSystemExit(
            "The resize action must be called with -a or --action."
        )

    client = _connect_client()
    server = client.servers.get_by_name(name)
    if server is None:
        return f"Instance {name} doesn't exist."

    # Check the configuration
    size = kwargs.get("size", None)
    if size is None:
        raise SaltCloudException("The new size is required")

    server_type = client.server_types.get_by_name(size)
    if server_type is None:
        raise SaltCloudException("The server size is not supported")

    __utils__["cloud.fire_event"](
        "event",
        "resizing instance",
        f"salt/cloud/{name}/resizing",
        args={"name": name},
        sock_dir=__opts__["sock_dir"],
        transport=__opts__["transport"],
    )

    node = show_instance(name, call="action")
    if node["state"] == "running":
        stop(name, call="action", wait=False)
        if not wait_until(name, "off"):
            return {"Error": f"Unable to resize {name}, command timed out"}

    server.change_type(server_type, kwargs.get("upgrade_disk", False))

    __utils__["cloud.fire_event"](
        "event",
        "resizing instance",
        f"salt/cloud/{name}/resized",
        args={"name": name},
        sock_dir=__opts__["sock_dir"],
        transport=__opts__["transport"],
    )

    return {"Resized": f"{name} was resized."}
